.pomodoro-list > .pomodoro-action-row {
    padding: 12px;
}

.pomodoro-action-name {
    margin-left: 6px;
    margin-bottom: 3px;
}

.pomodoro-action-keywords label {
    font-size: 9px;
    text-shadow: none;
    border-radius: 3px;
    padding: 3px;
    opacity: 0.6;
    background-color: alpha(@theme_fg_color, 0.08);
}
