/*
 * Copyright (c) 2011-2014 gnome-pomodoro contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

.extension-pomodoro-indicator .system-status-icon {
    padding: 0;
    margin: 0;
}

.extension-pomodoro-indicator-menu {
    max-width: 400px;
}

.extension-pomodoro-indicator-menu .state-item.active {
    font-weight: bold;
}

.extension-pomodoro-indicator-menu-toggle,
.extension-pomodoro-indicator-menu-timer {
    color: inherit;
    padding-top: 0;
    padding-bottom: 0;
}

.extension-pomodoro-indicator-menu-timer {
    margin-top: 5px;
    margin-bottom: 2px;
    spacing: 0;
}

.extension-pomodoro-indicator-menu-timer > StBoxLayout {
    margin-left: 10px;
    spacing: 4px;
}

.extension-pomodoro-indicator-menu-toggle .extension-pomodoro-indicator-menu-action {
    margin-left: 5px;
}

.extension-pomodoro-indicator-menu-timer-label {
    color: inherit;
    font-size: 1.6em;
    font-weight: 500;
    text-align: left;
    margin-left: 5px;
}

.extension-pomodoro-indicator-menu-timer-label-button {
    padding: 0;
    margin: 0;
    border: 0;
    background: none;
}

.extension-pomodoro-indicator-menu-action {
    border-radius: 18px;
    background-color: transparent;
}
.extension-pomodoro-indicator-menu-action,
.extension-pomodoro-indicator-menu-action:hover,
.extension-pomodoro-indicator-menu-action:focus,
.extension-pomodoro-indicator-menu-action:active {
    padding: 10px;
    border: 0px;
}

.extension-pomodoro-indicator-menu-action:hover,
.extension-pomodoro-indicator-menu-action:focus {
    background-color: #4e4e4e;
}

.extension-pomodoro-indicator-menu-action:active {
    background-color: #1e1e1e;
}

.message .extension-pomodoro-message-action {
    font-size: 10pt;
    font-weight: bold;
    padding: 10px 10px 10px 3px;
    border-radius: 0;
    background-color: transparent;
}

.message .extension-pomodoro-message-action:last-child {
    border-radius: 0 3px 3px 0;
    padding-right: 18px;
}

.extension-pomodoro-lightbox {
    background-color: rgba(0, 0, 0, 0.6);
}
.extension-pomodoro-lightbox-blurred {
    background-color: transparent;
}

.extension-pomodoro-dialog .extension-pomodoro-dialog-box {
    spacing: 5px;
    padding-bottom: 15px;
    width: 250pt;
}

.extension-pomodoro-dialog .extension-pomodoro-dialog-timer {
    font-size: 70pt;
    font-weight: normal;
    color: white;
}

.extension-pomodoro-dialog .extension-pomodoro-dialog-timer > StLabel:first-child {
    text-align: right;
}

.extension-pomodoro-dialog .extension-pomodoro-dialog-timer > StLabel:last-child {
    text-align: left;
}

.extension-pomodoro-dialog .extension-pomodoro-dialog-description {
    font-size: 13pt;
    font-weight: normal;
    color: white;
}
